import java.util.Arrays;
import java.util.NoSuchElementException;
import java.util.Random;

public class MinIntHeapExtended {

    private int[] heap;
    private int size;

    public MinIntHeapExtended() {
        heap = new int[10];
        size = 0;
    }

    public MinIntHeapExtended(int[] input, int size) {
        heap = input;
        this.size = size;
        // System.out.println(Arrays.toString(heap));
        // iterate over interior nodes from bottom to top, percolating each one down
        for (int i = size / 2; i > 0; i--) {
            int val = heap[i];
            // System.out.println("percolating " + val + " from " + i);
            int insertIndex = percolateDown(i, val);
            // System.out.println("to " + insertIndex);
            heap[insertIndex] = val;
            // System.out.println(Arrays.toString(heap));
        }
        // System.out.println(Arrays.toString(heap));
        checkHeap();
    }

    public int size() {
        return size;
    }

    public boolean isEmpty() {
        return size == 0;
    }

    private void grow() {
        checkHeap();
        int[] newHeap = new int[heap.length * 2];
        for (int i = 0; i < heap.length; i++) {
            newHeap[i] = heap[i];
        }
        heap = newHeap;
        checkHeap();
    }

    public void add(int val) {
        checkHeap();
        if (size == heap.length - 1) {
            grow();
        }
        size++;
        int insertIndex = percolateUp(size, val);
        heap[insertIndex] = val;
        checkHeap();
    }

    private int percolateUp(int hole, int val) {
        while (hole > 1 && val < heap[hole / 2]) {
            heap[hole] = heap[hole / 2];
            hole /= 2;
        }
        return hole;
    }

    public int removeMin() {
        if (isEmpty()) throw new NoSuchElementException();
        checkHeap();
        int ans = heap[1];
        int hole = percolateDown(1, heap[size]);
        heap[hole] = heap[size];
        size--;
        checkHeap();
        return ans;
    }

    private int percolateDown(int hole, int val) {
        while (hole <= size / 2) { // while hole is not a bottom-row leaf
            // find smaller child
            int left = hole * 2;
            int right = left + 1;
            int target = left;
            if (right <= size && heap[right] < heap[left]) {
                target = right;
            }
            // swap hole down the tree
            if (heap[target] < val) {
                heap[hole] = heap[target];
                hole = target;
            } else {
                break;
            }
        }
        return hole;
    }

    private void checkHeap() {
        // System.out.println(Arrays.toString(heap));
        // only need to iterate over interior nodes that can have children
        for (int i = 0; i <= size / 2; i++) {
            // System.out.println(heap[i] + " parent of " + heap[i * 2] + " and " + heap[i * 2 + 1]);
            if (heap[i] > heap[i * 2]) {
                throw new RuntimeException("index " + i + ":" + heap[i] + " > index " + i * 2 + ":" + heap[i * 2]
                        + " : " + Arrays.toString(heap));
            }
            if (i * 2 + 1 <= size && heap[i] > heap[i * 2 + 1]) {
                throw new RuntimeException("index " + i + ":" + heap[i]
                        + " > index " + (i * 2 + 1) + ":" + heap[i * 2 + 1] + " : " + Arrays.toString(heap));
            }
        }
    }

    public static void main(String[] args) {
        // MinIntHeapExtended testHeap = new MinIntHeapExtended();
        // Random rng = new Random();
        // while (testHeap.size() < 10) {
        //     testHeap.add(rng.nextInt(100));
        // }
        // while (testHeap.size() > 0) {
        //     System.out.print(testHeap.removeMin() + " ");
        // }
        // System.out.println();

        Random rng = new Random();
        int[] input = new int[100];
        for (int i = 1; i < input.length; i++) {
            input[i] = rng.nextInt(1000);
        }

        MinIntHeapExtended testHeap = new MinIntHeapExtended(input, input.length - 1);
        while (testHeap.size() > 0) {
            System.out.print(testHeap.removeMin() + " ");
        }
        System.out.println();
    }
}